import base64
import logging
import re
import six
import scrapy_rotated_proxy.signals as proxy_signals

from itertools import cycle
from scrapy.utils.misc import load_object
from twisted.internet import defer
from six.moves.urllib.parse import urlunparse
from six.moves.urllib.request import proxy_bypass
from six.moves.urllib.parse import unquote
from scrapy import signals
from scrapy.exceptions import NotConfigured, CloseSpider
from scrapy.http import Request
from scrapy.settings import Settings
from scrapy.utils.httpobj import urlparse_cached
from scrapy.utils.python import to_bytes
from scrapy_rotated_proxy.extensions import default_settings

if six.PY2:
    from urllib2 import _parse_proxy
else:
    from urllib.request import _parse_proxy

logger = logging.getLogger(__name__)


class RotatedProxy(object):
    @classmethod
    def from_crawler(cls, crawler):
        if not crawler.settings.getbool('ROTATED_PROXY_ENABLED'):
            raise NotConfigured

        auth_encoding = crawler.settings.get('HTTPPROXY_AUTH_ENCODING')
        o = cls(crawler, auth_encoding)
        crawler.signals.connect(o.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(o.spider_closed, signal=signals.spider_closed)
        crawler.signals.connect(o.proxy_block_received,
                                signal=proxy_signals.proxy_block)
        return o

    def __init__(self, crawler, auth_encoding='latin-1'):
        self.crawler = crawler
        self.auth_encoding = auth_encoding
        self.proxies_storage = load_object(
            crawler.settings.get('PROXY_STORAGE',
                                 getattr(default_settings, 'PROXY_STORAGE'))
            )(crawler.settings, self.auth_encoding)
        self.spider = None
        self.proxies = None
        self.black_proxies = {}
        self.proxy_gen = {}

    def spider_opened(self, spider):
        self.spider = spider
        self.proxies_storage.open_spider(spider)
        logger.info('Spider opened')

    def spider_closed(self, spider):
        self.proxies_storage.close_spider(spider)
        logger.info('Spider closed')

    def proxy_block_received(self, spider, response, exception):
        if response.meta.get('proxy'):
            scheme = urlparse_cached(response)[0]
            if scheme not in self.black_proxies:
                self.black_proxies.setdefault(scheme, set())
            if response.request.headers.get('Proxy-Authorization'):
                creds = response.request.headers.get('Proxy-Authorization')\
                        .split(b' ')[-1]
            else:
                creds = None
            self.black_proxies[scheme].add((creds, response.meta.get('proxy')))

            logger.info(
                'Block proxy: {proxy}, Total block {count} {scheme} proxy'.format(
                    proxy=response.meta.get('proxy'),
                    count=len(self.black_proxies[scheme]),
                    scheme=scheme
                ))

    @defer.inlineCallbacks
    def process_request(self, request, spider):
        if not self.proxies:
            self.proxies = yield self.proxies_storage.proxies()
            for scheme, proxies in self.proxies.items():
                logger.info(
                    'Loaded {count} {scheme} proxy from {origin}'.format(
                        count=len(proxies),
                        scheme=scheme,
                        origin=self.proxies_storage.__class__.__name__
                    ))

        # When Retry, dont_filter=True, reset proxy
        if 'proxy' in request.meta and not request.dont_filter:
            if request.meta['proxy'] is None:
                return
            creds, proxy_url = self._get_proxy(request.meta['proxy'], '')
            request.meta['proxy'] = proxy_url
            if creds and not request.headers.get('Proxy-Authorization'):
                request.headers['Proxy-Authorization'] = b'Basic ' + creds
            return
        elif not self.proxies:
            return

        parsed = urlparse_cached(request)
        scheme = parsed.scheme

        if scheme in ('http', 'https') and proxy_bypass(parsed.hostname):
            return

        if scheme in self.proxies:
            self._set_proxy(request, scheme)

    def _basic_auth_header(self, username, password):
        if six.PY2:
            username = username.encode(self.auth_encoding)
            password = password.encode(self.auth_encoding)

        user_pass = to_bytes(
            '{username}:{password}'.format(username=unquote(username),
                                           password=unquote(password)),
            encoding=self.auth_encoding
        )
        return base64.b64encode(user_pass).strip()

    def _set_proxy(self, request, scheme):
        creds, proxy = next(self._cycle_proxy(scheme))
        request.meta['proxy'] = proxy
        if creds:
            request.headers['Proxy-Authorization'] = b'Basic ' + creds

    def _cycle_proxy(self, scheme):
        if not self.proxy_gen.get(scheme):
            self.proxy_gen[scheme] = self._gen_proxy(scheme)
        return self.proxy_gen[scheme]

    def _gen_proxy(self, scheme):
        while True:
            self.proxies[scheme] -= self.black_proxies.get(scheme, set())
            if not self.proxies[scheme]:
                logger.info(
                    'Run out of all {scheme} proxies'.format(scheme=scheme))
                self.crawler.engine.close_spider(self.spider,
                                                 'Run out of All Proxy')
                break

            logger.debug('Left {count} {scheme} proxy to run'.format(
                count=len(self.proxies[scheme]),
                scheme=scheme
            ))
            for proxy_item in self.proxies[scheme]:
                if proxy_item in self.black_proxies.get(scheme, set()):
                    continue
                yield proxy_item

    def _get_proxy(self, url, orig_type=''):
        proxy_type, user, password, hostport = _parse_proxy(url)
        proxy_url = urlunparse(
            (proxy_type or orig_type, hostport, '', '', '', ''))

        creds = self._basic_auth_header(user, password) if user else None

        return creds, proxy_url



import logging

from scrapy.settings import Settings
from scrapy import Spider
from twisted.internet import defer
from txmongo.connection import ConnectionPool
from txmongo import filter as qf
from scrapy_rotated_proxy.extensions import default_settings


class BaseStorage(object):
    def __init__(self, settings, uri_key, db_key, coll_key, index_key):
        self.db_uri = settings.get(uri_key, getattr(default_settings, uri_key))
        self.db_name = settings.get(db_key, getattr(default_settings, db_key))
        self.coll_name = settings.get(coll_key,
                                      getattr(default_settings, coll_key))
        if index_key:
            self.db_index = settings.get(index_key,
                                         getattr(default_settings, index_key))
        self._db_client = None
        self._db = None
        self._coll = None
        self.logger = logging.getLogger(self.__class__.__name__)

    @defer.inlineCallbacks
    def open_spider(self, spider):
        self._db_client = yield ConnectionPool(self.db_uri)
        self._db = self._db_client[self.db_name]
        self._coll = self._db[self.coll_name]
        for index in self.db_index:
            yield self._coll.create_index(qf.sort(index))
        self.logger.info(
            'Spider opened: {spider_name}'.format(spider_name=spider.name,
                                                  storage_name=self.__class__))

    @defer.inlineCallbacks
    def close_spider(self, spider):
        yield self._db_client.disconnect()
        self.logger.info(
            'Spider closed: {spider_name}'.format(spider_name=spider.name,
                                                  storage_name=self.__class__))



from itertools import cycle

from twisted.internet import defer
from six.moves.urllib.parse import urlunparse
from twisted.internet.defer import returnValue

from scrapy_rotated_proxy import util
from scrapy_rotated_proxy.extensions.base_storage import BaseStorage


class MongoDBProxyStorage(BaseStorage):
    def __init__(self, settings, auth_encoding='latin-1'):
        super(MongoDBProxyStorage, self).__init__(
            settings,
            'PROXY_MONGODB_STORAGE_URI',
            'PROXY_MONGODB_STORAGE_DB',
            'PROXY_MONGODB_STORAGE_COLL',
            'PROXY_MONGODB_STORAGE_COLL_INDEX')
        self.auth_encoding = auth_encoding

    @defer.inlineCallbacks
    def proxies(self):
        '''
        :return: Dict
            {'http': [(UserPassBase64, proxy1_url), (None, proxy2_url)], 'https': []}
        '''

        def _gen_proxy(scheme, host, port, user=None, password=None):
            proxy_url = urlunparse((scheme, '{host}:{port}'.format(host=host,
                                                                   port=port),
                                    '', '', '', ''))
            creds = util._basic_auth_header(user, password,
                                            self.auth_encoding) if user else None
            return creds, proxy_url

        res = {}
        for scheme in ['http', 'https']:
            proxy_list = yield self._coll.find({'scheme': scheme})
            res[scheme] = set(map(lambda proxy: _gen_proxy(proxy['scheme'],
                                                           proxy['ip'],
                                                           proxy['port'],
                                                           proxy['username'],
                                                           proxy['password']),
                                  proxy_list))
        returnValue(res)




# DOWNLOADER_MIDDLEWARES = {}
# DOWNLOADER_MIDDLEWARES.update({
#     'scrapy.downloadermiddlewares.httpproxy.HttpProxyMiddleware': None,
#     'scrapy_rotated_proxy.downloadmiddlwares.proxy.RotatedProxy': 750,
# })

# default use Spider Settings Proxy
PROXY_STORAGE = 'scrapy_rotated_proxy.extensions.file_storage.FileProxyStorage'
PROXY_FILE_PATH = ''

# PROXY_STORAGE = 'scrapy_rotated_proxy.extensions.mongodb_storage.MongoDBProxyStorage'
PROXY_MONGODB_STORAGE_URI = 'mongodb://127.0.0.1:27017'
PROXY_MONGODB_STORAGE_DB = 'vps_management'
PROXY_MONGODB_STORAGE_COLL = 'service'
PROXY_MONGODB_STORAGE_COLL_INDEX = []


# settings_proxy.txt
# {
#     "HTTP_PROXIES": [
#         "http://host:ip"
#     ],
#     "HTTPS_PROXIES": [
#         "http://auth@host:ip"
#     ]
# }




import os
import sys
from functools import partial
from twisted.trial.unittest import TestCase, SkipTest
from twisted.internet import defer

from scrapy_rotated_proxy.downloadmiddlewares.proxy import RotatedProxy
from scrapy.downloadermiddlewares.httpproxy import HttpProxyMiddleware
from scrapy.exceptions import NotConfigured
from scrapy.http import Response, Request
from scrapy.spiders import Spider
from scrapy.crawler import Crawler
from scrapy.settings import Settings

spider = Spider('foo')


class TestDefaultHeadersMiddleware(TestCase):

    failureException = AssertionError

    def setUp(self):
        self._oldenv = os.environ.copy()

    def tearDown(self):
        os.environ = self._oldenv

    def test_not_enabled(self):
        settings = Settings({'ROTATEDPROXY_ENABLED': False})
        crawler = Crawler(spider, settings)
        self.assertRaises(NotConfigured, partial(RotatedProxy.from_crawler, crawler))

    @defer.inlineCallbacks
    def test_no_setting_proxies(self):
        crawler = Crawler(spider)
        mw = RotatedProxy(crawler)

        for url in ('http://e.com', 'https://e.com', 'file:///tmp/a'):
            req = Request(url)
            res = yield mw.process_request(req, spider)
            assert res is None
            self.assertEqual(req.url, url)
            self.assertEqual(req.meta, {})

    @defer.inlineCallbacks
    def test_setting_proxies(self):

        http_proxy = 'https://proxy.for.http:3128'
        https_proxy = 'http://proxy.for.https:8080'
        settings = Settings({'HTTP_PROXIES': [http_proxy], 'HTTPS_PROXIES': [https_proxy]})

        crawler = Crawler(spider, settings)
        mw = RotatedProxy(crawler)

        for url, proxy in [('http://e.com', http_proxy),
                ('https://e.com', https_proxy), ('file://tmp/a', None)]:
            req = Request(url)
            res = yield mw.process_request(req, spider)
            assert res is None
            self.assertEqual(req.url, url)
            self.assertEqual(req.meta.get('proxy'), proxy)

    @defer.inlineCallbacks
    def test_proxy_precedence_meta(self):
        settings = Settings({'HTTP_PROXIES': ['https://proxy.com']})
        crawler = Crawler(spider, settings)
        mw = RotatedProxy(crawler)
        req = Request('http://scrapytest.org', meta={'proxy': 'https://new.proxy:3128'})
        res = yield mw.process_request(req, spider)
        assert res is None
        self.assertEqual(req.meta, {'proxy': 'https://new.proxy:3128'})

    @defer.inlineCallbacks
    def test_proxy_auth(self):
        settings = Settings({'HTTP_PROXIES': ['https://user:pass@proxy:3128']})
        crawler = Crawler(spider, settings)
        mw = RotatedProxy(crawler)
        req = Request('http://scrapytest.org')
        res = yield mw.process_request(req, spider)
        assert res is None
        self.assertEqual(req.meta, {'proxy': 'https://proxy:3128'})
        self.assertEqual(req.headers.get('Proxy-Authorization'), b'Basic dXNlcjpwYXNz')
        # proxy from request.meta
        req = Request('http://scrapytest.org', meta={'proxy': 'https://username:password@proxy:3128'})
        res = yield mw.process_request(req, spider)
        assert res is None
        self.assertEqual(req.meta, {'proxy': 'https://proxy:3128'})
        self.assertEqual(req.headers.get('Proxy-Authorization'), b'Basic dXNlcm5hbWU6cGFzc3dvcmQ=')

    @defer.inlineCallbacks
    def test_proxy_auth_empty_passwd(self):
        settings = Settings({'HTTP_PROXIES': ['https://user:@proxy:3128']})
        crawler = Crawler(spider, settings)
        mw = RotatedProxy(crawler)
        req = Request('http://scrapytest.org')
        res = yield mw.process_request(req, spider)
        assert res is None
        self.assertEqual(req.meta, {'proxy': 'https://proxy:3128'})
        self.assertEqual(req.headers.get('Proxy-Authorization'), b'Basic dXNlcjo=')
        # proxy from request.meta
        req = Request('http://scrapytest.org', meta={'proxy': 'https://username:@proxy:3128'})
        res = yield mw.process_request(req, spider)
        assert res is None
        self.assertEqual(req.meta, {'proxy': 'https://proxy:3128'})
        self.assertEqual(req.headers.get('Proxy-Authorization'), b'Basic dXNlcm5hbWU6')

    @defer.inlineCallbacks
    def test_proxy_auth_encoding(self):
        # utf-8 encoding
        settings = Settings({'HTTP_PROXIES': [u'https://m\u00E1n:pass@proxy:3128']})
        crawler = Crawler(spider, settings)
        mw = RotatedProxy(crawler, auth_encoding='utf-8')
        req = Request('http://scrapytest.org')
        res = yield mw.process_request(req, spider)
        assert res is None
        self.assertEqual(req.meta, {'proxy': 'https://proxy:3128'})
        self.assertEqual(req.headers.get('Proxy-Authorization'), b'Basic bcOhbjpwYXNz')

        # proxy from request.meta
        req = Request('http://scrapytest.org', meta={'proxy': u'https://\u00FCser:pass@proxy:3128'})
        res = yield mw.process_request(req, spider)
        assert res is None
        self.assertEqual(req.meta, {'proxy': 'https://proxy:3128'})
        self.assertEqual(req.headers.get('Proxy-Authorization'), b'Basic w7xzZXI6cGFzcw==')

        # default latin-1 encoding
        mw = RotatedProxy(crawler, auth_encoding='latin-1')
        req = Request('http://scrapytest.org')
        res = yield mw.process_request(req, spider)
        assert res is None
        self.assertEqual(req.meta, {'proxy': 'https://proxy:3128'})
        self.assertEqual(req.headers.get('Proxy-Authorization'), b'Basic beFuOnBhc3M=')

        # proxy from request.meta, latin-1 encoding
        req = Request('http://scrapytest.org', meta={'proxy': u'https://\u00FCser:pass@proxy:3128'})
        res = yield mw.process_request(req, spider)
        assert res is None
        self.assertEqual(req.meta, {'proxy': 'https://proxy:3128'})
        self.assertEqual(req.headers.get('Proxy-Authorization'), b'Basic /HNlcjpwYXNz')

    @defer.inlineCallbacks
    def test_proxy_already_seted(self):
        settings = Settings({'HTTP_PROXIES': ['https://proxy.for.http:3128']})
        crawler = Crawler(spider, settings)
        mw = RotatedProxy(crawler)
        req = Request('http://noproxy.com', meta={'proxy': None})
        res = yield mw.process_request(req, spider)
        assert res is None
        assert 'proxy' in req.meta and req.meta['proxy'] is None


    @defer.inlineCallbacks
    def test_proxy_rotation(self):
        settings = Settings({'HTTP_PROXIES': ['https://proxy1.for.http:3128', 'https://proxy2.for.http:3128']})
        crawler = Crawler(spider, settings)
        mw = RotatedProxy(crawler)
        req1 = Request('http://scrapytest.org')
        req2 = Request('http://scrapytest.org')
        req3 = Request('http://scrapytest.org')
        res1 = yield mw.process_request(req1, spider)
        res2 = yield mw.process_request(req2, spider)
        res3 = yield mw.process_request(req3, spider)
        assert res1 is None
        assert res2 is None
        assert res3 is None
        res_list = [{'proxy': 'https://proxy1.for.http:3128'}, {'proxy': 'https://proxy2.for.http:3128'}]
        self.assertIn(req1.meta, res_list)
        self.assertIn(req2.meta, res_list)
        self.assertIn(req3.meta, res_list)
        self.assertNotEqual(req1.meta, req2.meta)
        self.assertEqual(req1.meta, req3.meta)

    @defer.inlineCallbacks
    def test_multi_scheme(self):
        settings = Settings({'HTTP_PROXIES': ['https://proxy1.for.http:3128'], 'HTTPS_PROXIES': ['https://proxy2.for.http:3128']})
        crawler = Crawler(spider, settings)
        mw = RotatedProxy(crawler)
        req_http = Request('http://scrapytest.org')
        req_https = Request('https://scrapytest.org')
        res_http = yield mw.process_request(req_http, spider)
        res_https = yield mw.process_request(req_https, spider)
        assert res_http is None
        assert res_https is None
        self.assertEqual(req_http.meta, {'proxy': 'https://proxy1.for.http:3128'})
        self.assertEqual(req_https.meta, {'proxy': 'https://proxy2.for.http:3128'})

